# coding: utf-8


def _jupyter_nbextension_paths():
    return [{
        'section': 'notebook',
        'src': 'nbextension/static',
        'dest': 'ipydagred3',
        'require': 'ipydagred3/extension'
    }]
